@@ -23,9 +23,10 @@ import (
 	. "github.com/onsi/ginkgo/extensions/table"
 	. "github.com/onsi/gomega"
 
+	ocpconfigv1 "github.com/openshift/api/config/v1"
 	routev1 "github.com/openshift/api/route/v1"
 	corev1 "k8s.io/api/core/v1"
-	extensionsv1beta1 "k8s.io/api/extensions/v1beta1"
+	networkingv1 "k8s.io/api/networking/v1"
 	storagev1 "k8s.io/api/storage/v1"
 	"k8s.io/apimachinery/pkg/api/resource"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -56,18 +57,37 @@ var _ = Describe("CDIConfig Controller reconcile loop", func() {
 	It("Should not update if no changes happened", func() {
 		reconciler, cdiConfig := createConfigReconciler(createConfigMap(operator.ConfigMapName, testNamespace))
 		err := reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
-		_, err = reconciler.Reconcile(reconcile.Request{})
+		Expect(err).ToNot(HaveOccurred())
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		Expect(err).ToNot(HaveOccurred())
 		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
 		Expect(err).ToNot(HaveOccurred())
 		// CDIConfig generated, now reconcile again without changes.
-		_, err = reconciler.Reconcile(reconcile.Request{})
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		Expect(err).ToNot(HaveOccurred())
 	})
 
+	It("Should update labels on CDIConfig when the ones on CR change", func() {
+		reconciler, cdiConfig := createConfigReconciler(createConfigMap(operator.ConfigMapName, testNamespace))
+		err := reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
+		Expect(err).ToNot(HaveOccurred())
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		// CDIConfig generated
+		reconciler.installerLabels[common.AppKubernetesPartOfLabel] = "new"
+		Expect(err).ToNot(HaveOccurred())
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
+		Expect(err).ToNot(HaveOccurred())
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(cdiConfig.Labels[common.AppKubernetesPartOfLabel]).To(Equal("new"))
+	})
+
 	DescribeTable("Should set proxyURL to override if no ingress or route exists", func(authority bool) {
 		reconciler, cdiConfig := createConfigReconciler(createConfigMap(operator.ConfigMapName, testNamespace))
-		_, err := reconciler.Reconcile(reconcile.Request{})
+		_, err := reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
 		Expect(err).ToNot(HaveOccurred())
 		override := "www.override-something.org.tt.test"
@@ -81,7 +101,7 @@ var _ = Describe("CDIConfig Controller reconcile loop", func() {
 		}
 		err = reconciler.client.Update(context.TODO(), cdi)
 		Expect(err).ToNot(HaveOccurred())
-		_, err = reconciler.Reconcile(reconcile.Request{})
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		Expect(err).ToNot(HaveOccurred())
 		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
 		Expect(err).ToNot(HaveOccurred())
@@ -101,7 +121,7 @@ var _ = Describe("CDIConfig Controller reconcile loop", func() {
 				*createIngress("test-ingress", "test-ns", testServiceName, testURL),
 			),
 		)
-		_, err := reconciler.Reconcile(reconcile.Request{})
+		_, err := reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
 		Expect(err).ToNot(HaveOccurred())
 		override := "www.override-something.org.tt.test"
@@ -115,7 +135,7 @@ var _ = Describe("CDIConfig Controller reconcile loop", func() {
 		}
 		err = reconciler.client.Update(context.TODO(), cdi)
 		Expect(err).ToNot(HaveOccurred())
-		_, err = reconciler.Reconcile(reconcile.Request{})
+		_, err = reconciler.Reconcile(context.TODO(), reconcile.Request{})
 		Expect(err).ToNot(HaveOccurred())
 		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
 		Expect(err).ToNot(HaveOccurred())
@@ -171,6 +191,22 @@ var _ = Describe("Controller ingress reconcile loop", func() {
 		Expect(err).ToNot(HaveOccurred())
 		Expect(*cdiConfig.Status.UploadProxyURL).To(Equal(testURL))
 	})
+
+	DescribeTable("Should not set proxyURL if invalid ingress exists", func(createIngress func(name, ns, service, url string) *networkingv1.Ingress) {
+		reconciler, cdiConfig := createConfigReconciler(createIngressList(
+			*createIngress("test-ingress", "test-ns", "service", testURL),
+		))
+		reconciler.uploadProxyServiceName = testServiceName
+		err := reconciler.reconcileIngress(cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(cdiConfig).ToNot(BeNil())
+		Expect(cdiConfig.Status).ToNot(BeNil())
+		Expect(cdiConfig.Status.UploadProxyURL).To(BeNil())
+	},
+		Entry("No default backend", createNoDefaultBackendIngress),
+		Entry("No service", createNoServiceIngress),
+		Entry("0 rules", createNoRulesIngress),
+	)
 })
 
 var _ = Describe("Controller route reconcile loop", func() {
@@ -283,6 +319,87 @@ var _ = Describe("Controller storage class reconcile loop", func() {
 	})
 })
 
+var _ = Describe("Controller ImportProxy reconcile loop", func() {
+	It("Should set ImportProxy to nil if no proxy configuration for import proxy exists", func() {
+		reconciler, cdiConfig := createConfigReconciler()
+		err := reconciler.reconcileRoute(cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(cdiConfig.Status.ImportProxy).To(BeNil())
+	})
+
+	var proxyHTTPURL = "http://user:pswd@www.myproxy.com"
+	var proxyHTTPSURL = "https://user:pswd@www.myproxy.com"
+	var noProxyDomains = ".myproxy.com,.noproxy.com"
+	var trustedCAProxy = "user-ca-bundle"
+
+	DescribeTable("Should set ImportProxy correctly if ClusterWideProxy with correct URLs exists", func(proxyHTTPURL string, proxyHTTPSURL string, noProxyDomains string, trustedCAName string, expect string, endpType string) {
+		reconciler, cdiConfig := createConfigReconciler(createClusterWideProxy(proxyHTTPURL, proxyHTTPSURL, noProxyDomains, trustedCAProxy))
+		_, err := reconciler.Reconcile(context.TODO(), reconcile.Request{})
+		Expect(err).ToNot(HaveOccurred())
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		switch endpType {
+		case common.ImportProxyHTTP:
+			Expect(*cdiConfig.Status.ImportProxy.HTTPProxy).To(Equal(expect))
+		case common.ImportProxyHTTPS:
+			Expect(*cdiConfig.Status.ImportProxy.HTTPSProxy).To(Equal(expect))
+		case common.ImportProxyNoProxy:
+			Expect(*cdiConfig.Status.ImportProxy.NoProxy).To(Equal(expect))
+		case common.ImportProxyConfigMapName:
+			Expect(*cdiConfig.Status.ImportProxy.TrustedCAProxy).To(Equal(expect))
+		default:
+		}
+	},
+		Entry("successfully get http proxy url", proxyHTTPURL, "", "", "", proxyHTTPURL, common.ImportProxyHTTP),
+		Entry("successfully get https proxy url", "", proxyHTTPSURL, "", "", proxyHTTPSURL, common.ImportProxyHTTPS),
+		Entry("successfully get the list of hostnames and/or CIDRs that proxy should not be used", "", "", noProxyDomains, "", noProxyDomains, common.ImportProxyNoProxy),
+		Entry("successfully get ConfiMap CA name", "", "", "", trustedCAProxy, trustedCAProxy, trustedCAProxy),
+	)
+
+	It("Should not change the CDIConfig when updating the ClusterWideProxy if the CDIConfig proxy information already exist", func() {
+		reconciler, cdiConfig := createConfigReconciler()
+		By("updating the CDIConfig with proxy information")
+		cdiConfig.Spec.ImportProxy = createImportProxy(proxyHTTPURL, proxyHTTPSURL, noProxyDomains, trustedCAProxy)
+		err := reconciler.reconcileImportProxy(cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+
+		By("creating cluster wide proxy")
+		proxy := createClusterWideProxy("http", "https", "noproxy", "ca")
+		err = reconciler.uncachedClient.Create(context.TODO(), proxy)
+		Expect(err).ToNot(HaveOccurred())
+		err = reconciler.reconcileImportProxy(cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(proxyHTTPURL).To(Equal(*cdiConfig.Status.ImportProxy.HTTPProxy))
+		Expect(proxyHTTPSURL).To(Equal(*cdiConfig.Status.ImportProxy.HTTPSProxy))
+		Expect(noProxyDomains).To(Equal(*cdiConfig.Status.ImportProxy.NoProxy))
+		Expect(trustedCAProxy).To(Equal(*cdiConfig.Status.ImportProxy.TrustedCAProxy))
+	})
+
+	It("Should create a new ConfigMap if ClusterWideProxy contains CA certificates name of an exiting ConfigMap in Openshift namespace", func() {
+		certificate := "ca-test"
+		reconciler, cdiConfig := createConfigReconciler(createClusterWideProxy(proxyHTTPURL, proxyHTTPSURL, noProxyDomains, ClusterWideProxyConfigMapName),
+			createClusterWideProxyCAConfigMap(certificate))
+		_, err := reconciler.Reconcile(context.TODO(), reconcile.Request{})
+		Expect(err).ToNot(HaveOccurred())
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: reconciler.configName}, cdiConfig)
+		Expect(err).ToNot(HaveOccurred())
+
+		test := &corev1.ConfigMap{}
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: ClusterWideProxyConfigMapName, Namespace: ClusterWideProxyConfigMapNameSpace}, test)
+
+		configMap := &corev1.ConfigMap{}
+		err = reconciler.client.Get(context.TODO(), types.NamespacedName{Name: common.ImportProxyConfigMapName, Namespace: reconciler.cdiNamespace}, configMap)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(configMap.Labels[common.AppKubernetesComponentLabel]).To(Equal("storage"))
+
+		cmCert, _ := configMap.Data[common.ImportProxyConfigMapKey]
+		Expect(string(cmCert)).To(Equal(certificate))
+	})
+})
+
 var _ = Describe("Controller create CDI config", func() {
 	It("Should return existing cdi config", func() {
 		reconciler, cdiConfig := createConfigReconciler()
@@ -322,25 +439,26 @@ var _ = Describe("Controller create CDI config", func() {
 		Expect(err).ToNot(HaveOccurred())
 		Expect(resConfig.Name).To(Equal("testconfig"))
 		Expect(resConfig.ObjectMeta.OwnerReferences[0].Name).To(Equal("testowner"))
+		Expect(resConfig.Labels[common.AppKubernetesPartOfLabel]).To(Equal(""))
 	})
 })
 
 var _ = Describe("getUrlFromIngress", func() {
 	//TODO: Once we get newer version of client-go, we need to switch to networking ingress.
 	It("Should return the url if backend service matches", func() {
-		ingress := &extensionsv1beta1.Ingress{
+		ingress := &networkingv1.Ingress{
 			TypeMeta: metav1.TypeMeta{
-				APIVersion: "extensions/v1beta1",
+				APIVersion: "networking/v1",
 			},
 			ObjectMeta: metav1.ObjectMeta{
 				Name:      "test",
 				Namespace: "test",
 			},
-			Spec: extensionsv1beta1.IngressSpec{
-				Backend: &extensionsv1beta1.IngressBackend{
-					ServiceName: testServiceName,
+			Spec: networkingv1.IngressSpec{
+				DefaultBackend: &networkingv1.IngressBackend{
+					Service: &networkingv1.IngressServiceBackend{Name: testServiceName},
 				},
-				Rules: []extensionsv1beta1.IngressRule{
+				Rules: []networkingv1.IngressRule{
 					{Host: testURL},
 				},
 			},
@@ -350,19 +468,19 @@ var _ = Describe("getUrlFromIngress", func() {
 	})
 
 	It("Should return blank if backend service does not match", func() {
-		ingress := &extensionsv1beta1.Ingress{
+		ingress := &networkingv1.Ingress{
 			TypeMeta: metav1.TypeMeta{
-				APIVersion: "extensions/v1beta1",
+				APIVersion: "networking/v1",
 			},
 			ObjectMeta: metav1.ObjectMeta{
 				Name:      "test",
 				Namespace: "test",
 			},
-			Spec: extensionsv1beta1.IngressSpec{
-				Backend: &extensionsv1beta1.IngressBackend{
-					ServiceName: testServiceName,
+			Spec: networkingv1.IngressSpec{
+				DefaultBackend: &networkingv1.IngressBackend{
+					Service: &networkingv1.IngressServiceBackend{Name: testServiceName},
 				},
-				Rules: []extensionsv1beta1.IngressRule{
+				Rules: []networkingv1.IngressRule{
 					{Host: testURL},
 				},
 			},
@@ -372,23 +490,23 @@ var _ = Describe("getUrlFromIngress", func() {
 	})
 
 	It("Should return the url if first rule backend service matches", func() {
-		ingress := &extensionsv1beta1.Ingress{
+		ingress := &networkingv1.Ingress{
 			TypeMeta: metav1.TypeMeta{
-				APIVersion: "extensions/v1beta1",
+				APIVersion: "networking/v1",
 			},
 			ObjectMeta: metav1.ObjectMeta{
 				Name:      "test",
 				Namespace: "test",
 			},
-			Spec: extensionsv1beta1.IngressSpec{
-				Rules: []extensionsv1beta1.IngressRule{
+			Spec: networkingv1.IngressSpec{
+				Rules: []networkingv1.IngressRule{
 					{
-						IngressRuleValue: extensionsv1beta1.IngressRuleValue{
-							HTTP: &extensionsv1beta1.HTTPIngressRuleValue{
-								Paths: []extensionsv1beta1.HTTPIngressPath{
+						IngressRuleValue: networkingv1.IngressRuleValue{
+							HTTP: &networkingv1.HTTPIngressRuleValue{
+								Paths: []networkingv1.HTTPIngressPath{
 									{
-										Backend: extensionsv1beta1.IngressBackend{
-											ServiceName: testServiceName,
+										Backend: networkingv1.IngressBackend{
+											Service: &networkingv1.IngressServiceBackend{Name: testServiceName},
 										},
 									},
 								},
@@ -404,38 +522,38 @@ var _ = Describe("getUrlFromIngress", func() {
 	})
 
 	It("Should return the url if any rule backend servicename matches", func() {
-		ingress := &extensionsv1beta1.Ingress{
+		ingress := &networkingv1.Ingress{
 			TypeMeta: metav1.TypeMeta{
-				APIVersion: "extensions/v1beta1",
+				APIVersion: "networking/v1",
 			},
 			ObjectMeta: metav1.ObjectMeta{
 				Name:      "test",
 				Namespace: "test",
 			},
-			Spec: extensionsv1beta1.IngressSpec{
-				Rules: []extensionsv1beta1.IngressRule{
+			Spec: networkingv1.IngressSpec{
+				Rules: []networkingv1.IngressRule{
 					{
-						IngressRuleValue: extensionsv1beta1.IngressRuleValue{
-							HTTP: &extensionsv1beta1.HTTPIngressRuleValue{
-								Paths: []extensionsv1beta1.HTTPIngressPath{
+						IngressRuleValue: networkingv1.IngressRuleValue{
+							HTTP: &networkingv1.HTTPIngressRuleValue{
+								Paths: []networkingv1.HTTPIngressPath{
 									{
-										Backend: extensionsv1beta1.IngressBackend{
-											ServiceName: "service1",
+										Backend: networkingv1.IngressBackend{
+											Service: &networkingv1.IngressServiceBackend{Name: "service1"},
 										},
 									},
 									{
-										Backend: extensionsv1beta1.IngressBackend{
-											ServiceName: "service2",
+										Backend: networkingv1.IngressBackend{
+											Service: &networkingv1.IngressServiceBackend{Name: "service2"},
 										},
 									},
 									{
-										Backend: extensionsv1beta1.IngressBackend{
-											ServiceName: testServiceName,
+										Backend: networkingv1.IngressBackend{
+											Service: &networkingv1.IngressServiceBackend{Name: testServiceName},
 										},
 									},
 									{
-										Backend: extensionsv1beta1.IngressBackend{
-											ServiceName: "service4",
+										Backend: networkingv1.IngressBackend{
+											Service: &networkingv1.IngressServiceBackend{Name: "service4"},
 										},
 									},
 								},
@@ -451,18 +569,18 @@ var _ = Describe("getUrlFromIngress", func() {
 	})
 
 	It("Should return blank if no http rule exists", func() {
-		ingress := &extensionsv1beta1.Ingress{
+		ingress := &networkingv1.Ingress{
 			TypeMeta: metav1.TypeMeta{
-				APIVersion: "extensions/v1beta1",
+				APIVersion: "networking/v1",
 			},
 			ObjectMeta: metav1.ObjectMeta{
 				Name:      "test",
 				Namespace: "test",
 			},
-			Spec: extensionsv1beta1.IngressSpec{
-				Rules: []extensionsv1beta1.IngressRule{
+			Spec: networkingv1.IngressSpec{
+				Rules: []networkingv1.IngressRule{
 					{
-						IngressRuleValue: extensionsv1beta1.IngressRuleValue{},
+						IngressRuleValue: networkingv1.IngressRuleValue{},
 					},
 				},
 			},
@@ -660,9 +778,10 @@ func createConfigReconciler(objects ...runtime.Object) (*CDIConfigReconciler, *c
 	// Register operator types with the runtime scheme.
 	s := scheme.Scheme
 	cdiv1.AddToScheme(s)
-	extensionsv1beta1.AddToScheme(s)
+	networkingv1.AddToScheme(s)
 	routev1.AddToScheme(s)
 	storagev1.AddToScheme(s)
+	ocpconfigv1.AddToScheme(s)
 
 	cdi := &cdiv1.CDI{
 		ObjectMeta: metav1.ObjectMeta{
@@ -687,6 +806,10 @@ func createConfigReconciler(objects ...runtime.Object) (*CDIConfigReconciler, *c
 		configName:             "cdiconfig",
 		cdiNamespace:           testNamespace,
 		uploadProxyServiceName: testServiceName,
+		installerLabels: map[string]string{
+			common.AppKubernetesPartOfLabel:  "",
+			common.AppKubernetesVersionLabel: "",
+		},
 	}
 	return r, cdiConfig
 }
@@ -730,34 +853,52 @@ func createRoute(name, ns, service string) *routev1.Route {
 	}
 }
 
-func createIngressList(ingresses ...extensionsv1beta1.Ingress) *extensionsv1beta1.IngressList {
-	list := &extensionsv1beta1.IngressList{
-		Items: []extensionsv1beta1.Ingress{},
+func createIngressList(ingresses ...networkingv1.Ingress) *networkingv1.IngressList {
+	list := &networkingv1.IngressList{
+		Items: []networkingv1.Ingress{},
 	}
 	list.Items = append(list.Items, ingresses...)
 	return list
 }
 
-func createIngress(name, ns, service, url string) *extensionsv1beta1.Ingress {
-	return &extensionsv1beta1.Ingress{
+func createIngress(name, ns, service, url string) *networkingv1.Ingress {
+	return &networkingv1.Ingress{
 		TypeMeta: metav1.TypeMeta{
-			APIVersion: "extensions/v1beta1",
+			APIVersion: "networking/v1",
 		},
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      name,
 			Namespace: ns,
 		},
-		Spec: extensionsv1beta1.IngressSpec{
-			Backend: &extensionsv1beta1.IngressBackend{
-				ServiceName: service,
+		Spec: networkingv1.IngressSpec{
+			DefaultBackend: &networkingv1.IngressBackend{
+				Service: &networkingv1.IngressServiceBackend{Name: service},
 			},
-			Rules: []extensionsv1beta1.IngressRule{
+			Rules: []networkingv1.IngressRule{
 				{Host: url},
 			},
 		},
 	}
 }
 
+func createNoDefaultBackendIngress(name, ns, service, url string) *networkingv1.Ingress {
+	res := createIngress(name, ns, service, url)
+	res.Spec.DefaultBackend = nil
+	return res
+}
+
+func createNoServiceIngress(name, ns, service, url string) *networkingv1.Ingress {
+	res := createIngress(name, ns, service, url)
+	res.Spec.DefaultBackend.Service = nil
+	return res
+}
+
+func createNoRulesIngress(name, ns, service, url string) *networkingv1.Ingress {
+	res := createIngress(name, ns, service, url)
+	res.Spec.Rules = []networkingv1.IngressRule{}
+	return res
+}
+
 func createConfigMap(name, namespace string) *corev1.ConfigMap {
 	isController := true
 	cm := &corev1.ConfigMap{
@@ -777,3 +918,13 @@ func createConfigMap(name, namespace string) *corev1.ConfigMap {
 	}
 	return cm
 }
+
+func createImportProxy(http, https, noproxy, ca string) *cdiv1.ImportProxy {
+	p := &cdiv1.ImportProxy{
+		HTTPProxy:      &http,
+		HTTPSProxy:     &https,
+		NoProxy:        &noproxy,
+		TrustedCAProxy: &ca,
+	}
+	return p
+}
