@@ -3,13 +3,19 @@ package controller
 import (
 	"context"
 	"crypto/rsa"
+	"encoding/json"
 	"fmt"
-	featuregates "kubevirt.io/containerized-data-importer/pkg/feature-gates"
+	"regexp"
+	"strconv"
 	"strings"
 
+	featuregates "kubevirt.io/containerized-data-importer/pkg/feature-gates"
+	"kubevirt.io/containerized-data-importer/pkg/token"
+
 	"github.com/go-logr/logr"
 	snapshotv1 "github.com/kubernetes-csi/external-snapshotter/v2/pkg/apis/volumesnapshot/v1beta1"
 	"github.com/pkg/errors"
+	corev1 "k8s.io/api/core/v1"
 	v1 "k8s.io/api/core/v1"
 	storagev1 "k8s.io/api/storage/v1"
 	extclientset "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
@@ -23,9 +29,12 @@ import (
 	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
 	cdiv1utils "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1/utils"
 	"kubevirt.io/containerized-data-importer/pkg/common"
+	"kubevirt.io/containerized-data-importer/pkg/util"
 	"kubevirt.io/containerized-data-importer/pkg/util/cert"
 	"kubevirt.io/containerized-data-importer/pkg/util/naming"
 	sdkapi "kubevirt.io/controller-lifecycle-operator-sdk/pkg/sdk/api"
+
+	ocpconfigv1 "github.com/openshift/api/config/v1"
 )
 
 const (
@@ -58,6 +67,10 @@ const (
 	AnnPopulatedFor = AnnAPIGroup + "/storage.populatedFor"
 	// AnnPrePopulated is a PVC annotation telling the datavolume controller that the PVC is already populated
 	AnnPrePopulated = AnnAPIGroup + "/storage.prePopulated"
+	// AnnPriorityClassName is PVC annotation to indicate the priority class name for importer, cloner and uploader pod
+	AnnPriorityClassName = AnnAPIGroup + "/storage.pod.priorityclassname"
+	// AnnPodRetainAfterCompletion is PVC annotation for retaining transfer pods after completion)
+	AnnPodRetainAfterCompletion = AnnAPIGroup + "/storage.pod.retainAfterCompletion"
 
 	// AnnPreviousCheckpoint provides a const to indicate the previous snapshot for a multistage import
 	AnnPreviousCheckpoint = AnnAPIGroup + "/storage.checkpoint.previous"
@@ -98,11 +111,33 @@ const (
 	// AnnImmediateBinding provides a const to indicate whether immediate binding should be performed on the PV (overrides global config)
 	AnnImmediateBinding = AnnAPIGroup + "/storage.bind.immediate.requested"
 
+	// AnnVddkVersion shows the last VDDK library version used by a DV's importer pod
+	AnnVddkVersion = AnnAPIGroup + "/storage.pod.vddk.version"
+	// AnnVddkHostConnection shows the last ESX host that serviced a DV's importer pod
+	AnnVddkHostConnection = AnnAPIGroup + "/storage.pod.vddk.host"
+
 	// PodRunningReason is const that defines the pod was started as a reason
 	podRunningReason = "Pod is running"
 
 	// Default value for preallocation option if not defined in DV or CDIConfig
 	defaultPreallocation = false
+
+	// ProxyCertVolName is the name of the volumecontaining certs
+	ProxyCertVolName = "cdi-proxy-cert-vol"
+	// ClusterWideProxyAPIGroup is the APIGroup for OpenShift Cluster Wide Proxy
+	ClusterWideProxyAPIGroup = "config.openshift.io"
+	// ClusterWideProxyAPIKind is the APIKind for OpenShift Cluster Wide Proxy
+	ClusterWideProxyAPIKind = "Proxy"
+	// ClusterWideProxyAPIVersion is the APIVersion for OpenShift Cluster Wide Proxy
+	ClusterWideProxyAPIVersion = "v1"
+	// ClusterWideProxyName is the OpenShift Cluster Wide Proxy object name. There is only one obj in the cluster.
+	ClusterWideProxyName = "cluster"
+	// ClusterWideProxyConfigMapName is the OpenShift Cluster Wide Proxy ConfigMap name for CA certificates.
+	ClusterWideProxyConfigMapName = "user-ca-bundle"
+	// ClusterWideProxyConfigMapNameSpace is the OpenShift Cluster Wide Proxy ConfigMap namespace for CA certificates.
+	ClusterWideProxyConfigMapNameSpace = "openshift-config"
+	// ClusterWideProxyConfigMapKey is the OpenShift Cluster Wide Proxy ConfigMap key name for CA certificates.
+	ClusterWideProxyConfigMapKey = "ca-bundle.crt"
 )
 
 const (
@@ -116,6 +151,18 @@ const (
 	AnnPodSidecarInjectionDefault = "false"
 )
 
+var (
+	vddkInfoMatch = regexp.MustCompile(`((.*; )|^)VDDK: (?P<info>{.*})`)
+)
+
+func isCrossNamespaceClone(dv *cdiv1.DataVolume) bool {
+	if dv.Spec.Source.PVC == nil {
+		return false
+	}
+
+	return dv.Spec.Source.PVC.Namespace != "" && dv.Spec.Source.PVC.Namespace != dv.Namespace
+}
+
 func checkPVC(pvc *v1.PersistentVolumeClaim, annotation string, log logr.Logger) bool {
 	// check if we have proper annotation
 	if !metav1.HasAnnotation(pvc.ObjectMeta, annotation) {
@@ -188,12 +235,23 @@ func getRequestedImageSize(pvc *v1.PersistentVolumeClaim) (string, error) {
 	return pvcSize.String(), nil
 }
 
-// returns the volumeMode which determines if the PVC is block PVC or not.
+// GetVolumeMode returns the volumeMode from PVC handling default empty value
+func GetVolumeMode(pvc *v1.PersistentVolumeClaim) v1.PersistentVolumeMode {
+	return getVolumeMode(pvc)
+}
+
+// returns the volumeMode from PVC handling default empty value
 func getVolumeMode(pvc *v1.PersistentVolumeClaim) v1.PersistentVolumeMode {
-	if pvc.Spec.VolumeMode != nil {
-		return *pvc.Spec.VolumeMode
+	return resolveVolumeMode(pvc.Spec.VolumeMode)
+}
+
+// resolveVolumeMode returns the volume mode if set, otherwise defaults to file system mode
+func resolveVolumeMode(volumeMode *v1.PersistentVolumeMode) v1.PersistentVolumeMode {
+	retVolumeMode := v1.PersistentVolumeFilesystem
+	if volumeMode != nil && *volumeMode == v1.PersistentVolumeBlock {
+		retVolumeMode = v1.PersistentVolumeBlock
 	}
-	return v1.PersistentVolumeFilesystem
+	return retVolumeMode
 }
 
 // checks if particular label exists in pvc
@@ -245,8 +303,9 @@ func newScratchPersistentVolumeClaimSpec(pvc *v1.PersistentVolumeClaim, pod *v1.
 }
 
 // CreateScratchPersistentVolumeClaim creates and returns a pointer to a scratch PVC which is created based on the passed-in pvc and storage class name.
-func CreateScratchPersistentVolumeClaim(client client.Client, pvc *v1.PersistentVolumeClaim, pod *v1.Pod, name, storageClassName string) (*v1.PersistentVolumeClaim, error) {
+func CreateScratchPersistentVolumeClaim(client client.Client, pvc *v1.PersistentVolumeClaim, pod *v1.Pod, name, storageClassName string, installerLabels map[string]string) (*v1.PersistentVolumeClaim, error) {
 	scratchPvcSpec := newScratchPersistentVolumeClaimSpec(pvc, pod, name, storageClassName)
+	util.SetRecommendedLabels(scratchPvcSpec, installerLabels, "cdi-controller")
 	if err := client.Create(context.TODO(), scratchPvcSpec); err != nil {
 		if !k8serrors.IsAlreadyExists(err) {
 			return nil, errors.Wrap(err, "scratch PVC API create errored")
@@ -294,11 +353,15 @@ func GetDefaultStorageClass(client client.Client) (*storagev1.StorageClass, erro
 
 // GetFilesystemOverhead determines the filesystem overhead defined in CDIConfig for this PVC's volumeMode and storageClass.
 func GetFilesystemOverhead(client client.Client, pvc *v1.PersistentVolumeClaim) (cdiv1.Percent, error) {
-	klog.V(1).Info("GetFilesystemOverhead with PVC", pvc)
 	if getVolumeMode(pvc) != v1.PersistentVolumeFilesystem {
 		return "0", nil
 	}
 
+	return GetFilesystemOverheadForStorageClass(client, pvc.Spec.StorageClassName)
+}
+
+// GetFilesystemOverheadForStorageClass determines the filesystem overhead defined in CDIConfig for the storageClass.
+func GetFilesystemOverheadForStorageClass(client client.Client, storageClassName *string) (cdiv1.Percent, error) {
 	cdiConfig := &cdiv1.CDIConfig{}
 	if err := client.Get(context.TODO(), types.NamespacedName{Name: common.ConfigName}, cdiConfig); err != nil {
 		if k8serrors.IsNotFound(err) {
@@ -309,28 +372,28 @@ func GetFilesystemOverhead(client client.Client, pvc *v1.PersistentVolumeClaim)
 		return "0", err
 	}
 
-	targetStorageClass, err := GetStorageClassByName(client, pvc.Spec.StorageClassName)
+	targetStorageClass, err := GetStorageClassByName(client, storageClassName)
 	if err != nil {
-		klog.V(1).Info("Storage class", pvc.Spec.StorageClassName, "not found, trying default storage class")
+		klog.V(3).Info("Storage class", storageClassName, "not found, trying default storage class")
 		targetStorageClass, err = GetStorageClassByName(client, nil)
 		if err != nil {
-			klog.V(1).Info("No default storage class found, continuing with global overhead")
+			klog.V(3).Info("No default storage class found, continuing with global overhead")
 			return cdiConfig.Status.FilesystemOverhead.Global, nil
 		}
 	}
 
-	klog.V(1).Info("target storage class for overhead", targetStorageClass)
-
 	if cdiConfig.Status.FilesystemOverhead == nil {
 		klog.Errorf("CDIConfig filesystemOverhead used before config controller ran reconcile. Hopefully this only happens during unit testing.")
 		return "0", nil
 	}
 
 	if targetStorageClass == nil {
-		klog.V(1).Info("Storage class", pvc.Spec.StorageClassName, "not found, continuing with global overhead")
+		klog.V(3).Info("Storage class", storageClassName, "not found, continuing with global overhead")
 		return cdiConfig.Status.FilesystemOverhead.Global, nil
 	}
 
+	klog.V(3).Info("target storage class for overhead", targetStorageClass.GetName())
+
 	perStorageConfig := cdiConfig.Status.FilesystemOverhead.StorageClass
 
 	storageClassOverhead, found := perStorageConfig[targetStorageClass.GetName()]
@@ -498,25 +561,75 @@ func podSucceededFromPVC(pvc *v1.PersistentVolumeClaim) bool {
 	return (podPhaseFromPVC(pvc) == v1.PodSucceeded)
 }
 
-func setConditionFromPodWithPrefix(anno map[string]string, prefix string, pod *v1.Pod) {
-	if pod.Status.ContainerStatuses != nil {
-		if pod.Status.ContainerStatuses[0].State.Running != nil {
-			anno[prefix] = "true"
-			anno[prefix+".message"] = ""
-			anno[prefix+".reason"] = podRunningReason
-		} else {
-			anno[AnnRunningCondition] = "false"
-			if pod.Status.ContainerStatuses[0].State.Waiting != nil {
-				anno[prefix+".message"] = pod.Status.ContainerStatuses[0].State.Waiting.Message
-				anno[prefix+".reason"] = pod.Status.ContainerStatuses[0].State.Waiting.Reason
-			} else if pod.Status.ContainerStatuses[0].State.Terminated != nil {
-				anno[prefix+".message"] = pod.Status.ContainerStatuses[0].State.Terminated.Message
-				anno[prefix+".reason"] = pod.Status.ContainerStatuses[0].State.Terminated.Reason
+func setAnnotationsFromPodWithPrefix(anno map[string]string, pod *v1.Pod, prefix string) {
+	if pod == nil || pod.Status.ContainerStatuses == nil {
+		return
+	}
+	annPodRestarts, _ := strconv.Atoi(anno[AnnPodRestarts])
+	podRestarts := int(pod.Status.ContainerStatuses[0].RestartCount)
+	if podRestarts >= annPodRestarts {
+		anno[AnnPodRestarts] = strconv.Itoa(podRestarts)
+	}
+	setVddkAnnotations(anno, pod)
+	containerState := pod.Status.ContainerStatuses[0].State
+	if containerState.Running != nil {
+		anno[prefix] = "true"
+		anno[prefix+".message"] = ""
+		anno[prefix+".reason"] = podRunningReason
+	} else {
+		anno[AnnRunningCondition] = "false"
+		if containerState.Waiting != nil && containerState.Waiting.Reason != "CrashLoopBackOff" {
+			anno[prefix+".message"] = simplifyKnownMessage(containerState.Waiting.Message)
+			anno[prefix+".reason"] = containerState.Waiting.Reason
+		} else if containerState.Terminated != nil {
+			anno[prefix+".message"] = simplifyKnownMessage(containerState.Terminated.Message)
+			anno[prefix+".reason"] = containerState.Terminated.Reason
+			if strings.Contains(containerState.Terminated.Message, common.PreallocationApplied) {
+				anno[AnnPreallocationApplied] = "true"
 			}
 		}
 	}
 }
 
+func simplifyKnownMessage(msg string) string {
+	if strings.Contains(msg, "is larger than the reported available") ||
+		strings.Contains(msg, "no space left on device") ||
+		strings.Contains(msg, "file largest block is bigger than maxblock") {
+		return "DataVolume too small to contain image"
+	}
+
+	return msg
+}
+
+func setVddkAnnotations(anno map[string]string, pod *corev1.Pod) {
+	if pod.Status.ContainerStatuses[0].State.Terminated == nil {
+		return
+	}
+	terminationMessage := pod.Status.ContainerStatuses[0].State.Terminated.Message
+	klog.V(1).Info("Saving VDDK annotations from pod status message: ", "message", terminationMessage)
+
+	var terminationInfo string
+	matches := vddkInfoMatch.FindAllStringSubmatch(terminationMessage, -1)
+	for index, matchName := range vddkInfoMatch.SubexpNames() {
+		if matchName == "info" && len(matches) > 0 {
+			terminationInfo = matches[0][index]
+			break
+		}
+	}
+
+	var vddkInfo util.VddkInfo
+	err := json.Unmarshal([]byte(terminationInfo), &vddkInfo)
+	if err != nil {
+		return
+	}
+	if vddkInfo.Host != "" {
+		anno[AnnVddkHostConnection] = vddkInfo.Host
+	}
+	if vddkInfo.Version != "" {
+		anno[AnnVddkVersion] = vddkInfo.Version
+	}
+}
+
 func setBoundConditionFromPVC(anno map[string]string, prefix string, pvc *v1.PersistentVolumeClaim) {
 	switch pvc.Status.Phase {
 	case v1.ClaimBound:
@@ -552,7 +665,8 @@ func createScratchNameFromPvc(pvc *v1.PersistentVolumeClaim) string {
 	return naming.GetResourceName(pvc.Name, common.ScratchNameSuffix)
 }
 
-func getPodsUsingPVCs(c client.Client, namespace string, names sets.String, allowReadOnly bool) ([]v1.Pod, error) {
+// GetPodsUsingPVCs returns Pods currently using PVCs
+func GetPodsUsingPVCs(c client.Client, namespace string, names sets.String, allowReadOnly bool) ([]v1.Pod, error) {
 	pl := &v1.PodList{}
 	// hopefully using cached client here
 	err := c.List(context.TODO(), pl, &client.ListOptions{Namespace: namespace})
@@ -663,27 +777,147 @@ func GetPreallocation(client client.Client, dataVolume *cdiv1.DataVolume) bool {
 	return cdiconfig.Status.Preallocation
 }
 
-// GetStorageClassNameForDV returns storage class to be used for the DV's PVC
-func GetStorageClassNameForDV(c client.Client, dv *cdiv1.DataVolume) string {
-	// If DV has a SC, return it
-	if dv != nil && dv.Spec.PVC != nil && dv.Spec.PVC.StorageClassName != nil && *dv.Spec.PVC.StorageClassName != "" {
-		return *dv.Spec.PVC.StorageClassName
+// GetClusterWideProxy returns the OpenShift cluster wide proxy object
+func GetClusterWideProxy(r client.Client) (*ocpconfigv1.Proxy, error) {
+	clusterWideProxy := &ocpconfigv1.Proxy{}
+	// Ignore both no CRD found (IgnoreIsNoMatch) and the object itself not existing IsNotFound because we want to skip if not
+	// in Open Shift.
+	if err := r.Get(context.TODO(), types.NamespacedName{Name: ClusterWideProxyName}, clusterWideProxy); IgnoreIsNoMatchError(err) != nil && !k8serrors.IsNotFound(err) {
+		return nil, err
+	}
+	return clusterWideProxy, nil
+}
+
+// GetImportProxyConfig attempts to import proxy URLs if configured in the CDIConfig.
+func GetImportProxyConfig(config *cdiv1.CDIConfig, field string) (string, error) {
+	if config == nil {
+		return "", errors.Errorf("failed to get field, the CDIConfig is nil\n")
+	}
+	if config.Status.ImportProxy == nil {
+		return "", errors.Errorf("failed to get field, the CDIConfig ImportProxy is nil\n")
 	}
 
-	// If DV's PVC has a SC, return it
-	pvc := &v1.PersistentVolumeClaim{}
-	// TODO change when PVC's name is different from DV's name
-	err := c.Get(context.TODO(), types.NamespacedName{Name: dv.Name, Namespace: dv.Namespace}, pvc)
-	if err == nil && pvc.Spec.StorageClassName != nil {
-		return *pvc.Spec.StorageClassName
-	}
-
-	// If there is a default SC, return it
-	sc, _ := GetDefaultStorageClass(c)
-	if sc != nil {
-		return sc.Name
+	switch field {
+	case common.ImportProxyHTTP:
+		if config.Status.ImportProxy.HTTPProxy != nil {
+			return *config.Status.ImportProxy.HTTPProxy, nil
+		}
+	case common.ImportProxyHTTPS:
+		if config.Status.ImportProxy.HTTPSProxy != nil {
+			return *config.Status.ImportProxy.HTTPSProxy, nil
+		}
+	case common.ImportProxyNoProxy:
+		if config.Status.ImportProxy.NoProxy != nil {
+			return *config.Status.ImportProxy.NoProxy, nil
+		}
+	case common.ImportProxyConfigMapName:
+		if config.Status.ImportProxy.TrustedCAProxy != nil {
+			return *config.Status.ImportProxy.TrustedCAProxy, nil
+		}
+	default:
+		return "", errors.Errorf("CDIConfig ImportProxy does not have the field: %s\n", field)
 	}
 
 	// If everything fails, return blank
-	return ""
+	return "", nil
+}
+
+func getPriorityClass(pvc *v1.PersistentVolumeClaim) string {
+	anno := pvc.GetAnnotations()
+	return anno[AnnPriorityClassName]
+}
+
+func shouldDeletePod(pvc *v1.PersistentVolumeClaim) bool {
+	return pvc.GetAnnotations()[AnnPodRetainAfterCompletion] != "true" || pvc.GetAnnotations()[AnnRequiresScratch] == "true" || pvc.DeletionTimestamp != nil
+}
+
+// AddFinalizer adds a finalizer to a resource
+func AddFinalizer(obj metav1.Object, name string) {
+	if HasFinalizer(obj, name) {
+		return
+	}
+
+	obj.SetFinalizers(append(obj.GetFinalizers(), name))
+}
+
+// RemoveFinalizer removes a finalizer from a resource
+func RemoveFinalizer(obj metav1.Object, name string) {
+	if !HasFinalizer(obj, name) {
+		return
+	}
+
+	var finalizers []string
+	for _, f := range obj.GetFinalizers() {
+		if f != name {
+			finalizers = append(finalizers, f)
+		}
+	}
+
+	obj.SetFinalizers(finalizers)
+}
+
+// HasFinalizer returns true if a resource has a specific finalizer
+func HasFinalizer(object metav1.Object, value string) bool {
+	for _, f := range object.GetFinalizers() {
+		if f == value {
+			return true
+		}
+	}
+	return false
+}
+
+func validateCloneTokenPVC(validator token.Validator, source, target *corev1.PersistentVolumeClaim) error {
+	if source.Namespace == target.Namespace {
+		return nil
+	}
+
+	tok, ok := target.Annotations[AnnCloneToken]
+	if !ok {
+		return errors.New("clone token missing")
+	}
+
+	tokenData, err := validator.Validate(tok)
+	if err != nil {
+		return errors.Wrap(err, "error verifying token")
+	}
+
+	return validateTokenData(tokenData, source.Namespace, source.Name, target.Namespace, target.Name)
+}
+
+func validateCloneTokenDV(validator token.Validator, dv *cdiv1.DataVolume) error {
+	if dv.Spec.Source.PVC == nil || dv.Spec.Source.PVC.Namespace == "" || dv.Spec.Source.PVC.Namespace == dv.Namespace {
+		return nil
+	}
+
+	tok, ok := dv.Annotations[AnnCloneToken]
+	if !ok {
+		return errors.New("clone token missing")
+	}
+
+	tokenData, err := validator.Validate(tok)
+	if err != nil {
+		return errors.Wrap(err, "error verifying token")
+	}
+
+	return validateTokenData(tokenData, dv.Spec.Source.PVC.Namespace, dv.Spec.Source.PVC.Name, dv.Namespace, dv.Name)
+}
+
+func validateTokenData(tokenData *token.Payload, srcNamespace, srcName, targetNamespace, targetName string) error {
+	if tokenData.Operation != token.OperationClone ||
+		tokenData.Name != srcName ||
+		tokenData.Namespace != srcNamespace ||
+		tokenData.Resource.Resource != "persistentvolumeclaims" ||
+		tokenData.Params["targetNamespace"] != targetNamespace ||
+		tokenData.Params["targetName"] != targetName {
+		return errors.New("invalid token")
+	}
+
+	return nil
+}
+
+func addAnnotation(obj metav1.Object, key, value string) {
+	if obj.GetAnnotations() == nil {
+		obj.SetAnnotations(make(map[string]string))
+	}
+	obj.GetAnnotations()[key] = value
 }
